-------------------------------------------------------------------------------
-- Mission: for the lecture notes on VHDL
--    illustrate entity declaration and architecture body,
--    show usage of generics, oppose procedural and structural styles.
-- Functionality: binary code to Gray code converter, also see 
--    usefulstuff.vhd for conversion functions and graycnt.vhd for counter.
--    Example designed to feature parametrized word width and delay,
--    as well as if and for generate statements.
-- Author: H.Kaeslin.
-------------------------------------------------------------------------------
-- Modification history:
-- 11.9.97: created as a didactically better replacement for [vhd]addsub.
-- 22.4.98: copied to Synopsys but not adapted yet. may want to implement 
--          conversion with functions, see usefulstuff.vhd and puzzle2inv.vhd.
-- 4.6.98: ported to Synopsys with major modifications relating to data types.
--         - natural is not supported but gets automatically interpreted
--           as integer (warning VHDL-2023).
--         - time is not supported (error VHDL-2254) and had to be withdrawn.
-- 18.2.14: adapted to revised signal naming convention.
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
library GTECH;
use GTECH.GTECH_components.all;   -- generic component library by Synopsys
-------------------------------------------------------------------------------
--
-- bit positions           4     3     2     1     0          X = XOR
--
--                         |     |     |     |     |
-- binary to Gray          |--.  |--.  |--.  |--.  |          # of X-ops on
-- conversion              |  `--X  `--X  `--X  `--X          longest paths
-- for width=5             |     |     |     |     |          = 1
--                         v     v     v     v     v
--
-------------------------------------------------------------------------------

entity binary2gray is
   generic (width : natural := 5);    -- word width with default value
--          tpd : time := 2.7 ns );   -- propagation delay
   port (Inp_DI : in std_logic_vector(width-1 downto 0);     -- binary in
         Oup_DO : out std_logic_vector(width-1 downto 0) );  -- Gray out
end binary2gray;

--=============================================================================

---- mapping specified by procedural computation
architecture behavioral of binary2gray is

   -- conversion formulated as in-place computation on a scratchpad variable
   function bintogray (arg : std_logic_vector) return std_logic_vector is 
      variable scrapa : std_logic_vector(arg'length-1 downto 0); 
   begin
      scrapa := arg;
      for i in 0 to arg'length-2 loop      -- note MSB remains unchanged
         if scrapa(i+1)='1' then
            scrapa(i) := not scrapa(i);
         end if;
      end loop;
      return scrapa;      
   end bintogray;

begin
   
      Oup_DO <= bintogray(Inp_DI);   -- after tpd;
      
end behavioral;

--=============================================================================

---- logic network described as a bunch of interconnected logic gates.
---- though simple and effective, the problem with this architecture is
---- that it requires no internal signals and no conditional generate,
---- two features we would like to demonstrate for teaching purposes.
architecture structural_toosimple of binary2gray is

   -- component declarations
   -- redundant as GTECH library has been made available at the top

   -- signal declarations
   -- no internal signals required for this example

begin

   -- assemble all bit slices by instantiating logic gates
   lower_slices : for i in 0 to width-2 generate
      uxo : GTECH_XOR2
         port map (A => Inp_DI(i), B => Inp_DI(i+1), Z => Oup_DO(i));
   end generate lower_slices;
   uni : GTECH_BUF
      port map (A => Inp_DI(width-1), Z => Oup_DO(width-1));

end structural_toosimple;

--=============================================================================

---- logic network described as a bunch of interconnected logic gates
architecture structural of binary2gray is
   
   -- component declarations
   -- redundant as explained before but kept here for teaching purposes
   component GTECH_XNOR2   -- 2-inp xnor gate
      port (A, B : in std_logic;
            Z : out std_logic);
   end component;
   component GTECH_NOT     -- inverter
      port (A : in std_logic;
            Z : out std_logic);
   end component;

   -- signal declarations
   signal Node_D : std_logic_vector(width-1 downto 0);  -- internal nodes

begin

   -- assemble all bit slices by instantiating logic gates
   any_slice : for i in 0 to width-1 generate  -- repetitive elaboration
      -- a row of xnors except for the MSB which requires an inverter
      less_significant : if i<width-1 generate  -- conditional elaboration
         uxn : GTECH_XNOR2  -- component instantiation
             port map (A=>Inp_DI(i), B=>Inp_DI(i+1), Z=>Node_D(i));
      end generate;
      most_significant : if i=width-1 generate  -- conditional elaboration
         uin : GTECH_NOT  -- component instantiation
            port map (A=>Inp_DI(i), Z=>Node_D(i));
      end generate;
      -- a final row of inverters
      uin : GTECH_NOT  -- component instantiation
         port map (A=>Node_D(i), Z=>Oup_DO(i));
   end generate;

end structural;

--=============================================================================
