// Mission: Illustrate how to model a Medvedev machine in two processes.
//    Example designed to demonstrate assertion-based verification, the usage 
//    of functions from a package, and the absence of any type conversions. 
// Functionality: w-bit Gray counter with enable and asynchronous reset. 
//    Operates with double conversion: Gray->binary, increment, binary->Gray,
//    which is not necessarily the most economic nor the fastest solution. 
// Author: H.Kaeslin 
//-----------------------------------------------------------------------------

`include "../sourcecode/graydefs.sv"    // the word width to be used  
import grayconvPkg::*;   // my own set of Gray code converter functions

module graycnt 
   ( input logic Clk_CI, Rst_RBI, Ena_SI, 
     output logic [`GRAYWIDTH-1:0] Count_DO );

   // present state and next state
   logic [`GRAYWIDTH-1:0] Count_DN, Count_DP;

   // computation of next state
   assign Count_DN = bin2gray(gray2bin(Count_DP) + 1); 

   // updating of state
   always_ff @(posedge Clk_CI, negedge Rst_RBI)
       if (~Rst_RBI)                 // if active low
         Count_DP <= 0;              // reset to all zeroes
       else                          // if rising clock
          if (Ena_SI)                // and enable is 1 
             Count_DP <= Count_DN;   // make next state the present one

   // assignment of state to output port
   assign Count_DO = Count_DP; // output assignment

//-------------------------------------------------------------------------
// set of assertions for addressing the various functional mechanisms 

  // check asynchronous reset mechanism
   a_async_reset: assert property ( 
                     @(posedge Rst_RBI) 
                     Count_DO == '{default:1'b0} ) 
                     else
      $error("Non-zero output immediately after removal of async. reset.");
   // or alternatively:   $countones(Count_DO) == 0 );

   // check enable/disable mechanism
   a_enable_disable: assert property ( 
                     @(posedge Clk_CI) disable iff (~Rst_RBI) 
                     ~Ena_SI |=> Count_DO == $past(Count_DO) ) 
                     else
      $error("Output has changed in spite of counting being disabled.");

   // check unit distance counting steps 
   a_unit_distance_step: assert property ( 
      // sampling time vvvvvvvvvvvvvv     criterion vvvvvvvv of exclusion  
                     @(posedge Clk_CI) disable iff (~Rst_RBI) 
      //precondition vvvvvv     vvvvvvvvv property to be checked vvvvvvvvvv 
                     Ena_SI |=> $countones(Count_DO ^ $past(Count_DO)) == 1 ) 
      // optional pass action block (empty in this example) 
                     else
      // optional fail action block
      $error("Consecutive output patterns differ in 0 or more than 1 bit.");

endmodule
