-- Mission: Illustrate how to model a Medvedev machine in two processes.
--    Example designed to demonstrate the usage of functions from a package, 
--    of IEEE numeric_std, type conversions, and a parametrized word width. 
-- Functionality: w-bit Gray counter with enable and asynchronous reset. 
--    Operates with double conversion: Gray->binary, increment, binary->Gray,
--    which is not necessarily the most economic nor the fastest solution. 
-- Author: H.Kaeslin.
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.grayconv.all;          -- my own set of Gray code converter functions
-------------------------------------------------------------------------------

entity graycnt is
   generic (
      GRAYWIDTH : integer := 5 );   -- default value for number of state bits
   port (
      Clk_CI   : in  std_logic;
      Rst_RBI  : in  std_logic;
      Ena_SI   : in  std_logic;
      Count_DO : out std_logic_vector((GRAYWIDTH-1) downto 0) );
end graycnt;

-------------------------------------------------------------------------------

architecture computed_state of graycnt is

   -- present state and next state
   signal Count_DP, Count_DN : std_logic_vector(GRAYWIDTH-1 downto 0);

begin

   ---- computation of next state
   Count_DN <= bintogray(std_logic_vector(
      unsigned(graytobin(Count_DP)) + 1 ));

   ---- updating of state
   p_memzing : process (Clk_CI,Rst_RBI)
   begin
      -- activities triggered by asynchronous reset
      if Rst_RBI='0' then
         Count_DP <= (others => '0');   -- width-independent shorthand
      -- activities triggered by rising edge of clock
      elsif Clk_CI'event and Clk_CI='1' then
         -- proceed to next state only if enable is asserted
         if Ena_SI='1' then
            Count_DP <= Count_DN;
         end if;
      end if;
   end process p_memzing;

   ---- assignment of state to output only signal
   Count_DO <= Count_DP;

end computed_state;
