-- Mission: illustrate the use of a package in VHDL.
-- Functionality: Gray code <-> binary code conversion functions.
-- Author: H.Kaeslin.
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
-------------------------------------------------------------------------------

-- package declaration
package grayconv is
   function bintogray (arg : std_logic_vector) return std_logic_vector;
   function graytobin (arg : std_logic_vector) return std_logic_vector;
end grayconv;

-------------------------------------------------------------------------------
--
--                         |     |     |     |     |
-- binary to Gray          |--.  |--.  |--.  |--.  |          # of X-ops on
-- conversion              |  `--X  `--X  `--X  `--X          longest paths
-- for GRAYWIDTH=5         |     |     |     |     |          = 1
--                         v     v     v     v     v
-- bit positions           4     3     2     1     0          X = XOR
--                         |     |     |     |     |
-- Gray to binary          |  ,--X  ,--X  ,--X  ,--X          # of X-ops on
-- conversion              |--'  |--'  |--'  |--'  |          longest path
-- for GRAYWIDTH=5         |     |     |     |     |          = GRAYWIDTH-1
--                         v     v     v     v     v
--
-------------------------------------------------------------------------------

-- package body
package body grayconv is

   -- purpose: converts binary code into Gray code
   -- by way of in-place computation on a scratchpad variable
   function bintogray (arg : std_logic_vector) return std_logic_vector is 
      variable scrapa : std_logic_vector(arg'length-1 downto 0); 
   begin
      scrapa := arg;
      for i in 0 to arg'length-2 loop       -- MSB remains unchanged
         if scrapa(i+1)='1' then
            scrapa(i) := not scrapa(i);
         end if;
      end loop;
      return scrapa;      
   end bintogray;

   -- purpose: converts Gray code into binary code
   -- by way of in-place computation on a scratchpad variable
   function graytobin (arg : std_logic_vector) return std_logic_vector is 
      variable scrapa : std_logic_vector(arg'length-1 downto 0); 
   begin
      scrapa := arg;
      for i in arg'length-2 downto 0 loop   -- MSB remains unchanged
         if scrapa(i+1)='1' then
            scrapa(i) := not scrapa(i);
         end if;
      end loop;
      return scrapa;      
   end graytobin;

end grayconv;
