// Mission: Illustrate how to model a Mealy machine with two processes.
//    Example designed to include an asynchronous reset, three-state outputs,
//    self loops, symbolic encoding of states and outputs, plus handling of
//    parasitic states and inputs so as to eliminate any chance for lockup.
// Functionality: See state graph below.     
// Author: B.Muheim.
// ----------------------------------------------------------------------------

module mealy5st
   ( input  logic Clk_CI,
     input  logic Rst_RBI,
     input  logic [1:0] Inp_DI,
     output logic [1:0] Oup_DO ); // ternary, 11 not used

//-----------------------------------------------------------------------------
// Inp_DI / Oup_DO                                                   (State_DP)
//
//              00 /        00 /        00 /        00 /        00 /
//             EMPTY        IBTW        IBTW        IBTW        FULL
//               _           _           _           _           _
//              | |  01 /   | |  01 /   | |  01 /   | |  01 /   | |
//              | |  IBTW   | |  IBTW   | |  IBTW   | |  FULL   | |
//         ---> v / ------> v / ------> v / ------> v / ------> v / ----
//   10 / |     (0)         (1)         (2)         (3)         (4)     | 01 /
//  EMPTY  -----   <--------   <-------- ^ <--------   <--------   <----  FULL
//                    10 /        10 /   |    10 /        10 / 
//                   EMPTY        IBTW   |    IBTW        IBTW              
//                                       o
//-----------------------------------------------------------------------------

   typedef enum integer {ST0, ST1, ST2, ST3, ST4} state_type; 

   // symbolic encodings of output
   const logic [1:0] EMPTY = 'b10;   // empty
   const logic [1:0] IBTW  = 'b00;   // in between
   const logic [1:0] FULL  = 'b01;   // full
   const logic [1:0] HIMP  = 'bZZ;   // high-impedance

   // present state and next state
   state_type State_DP, State_DN;

   // computation of next state and present outputs
   // ------------------------------------------------------------------------
   always_comb  begin
      // default assignments
      State_DN = State_DP;   // remain in present state 
      Oup_DO = IBTW;         // output "in between"
      // nondefault transitions and outputs
      case (State_DP)
         ST0 :
            if      (Inp_DI=='b00) Oup_DO   = EMPTY;
            else if (Inp_DI=='b10) Oup_DO   = EMPTY;
            else if (Inp_DI=='b01) State_DN = ST1;
            else Oup_DO   = EMPTY;   // parasitic input 11, treat as 00
         ST1 :
            if (Inp_DI=='b10) begin
              Oup_DO   = EMPTY;
              State_DN = ST0;
            end
            else if (Inp_DI=='b01) State_DN = ST2;
         ST2 :
            if      (Inp_DI=='b10) State_DN = ST1;
            else if (Inp_DI=='b01) State_DN = ST3;
         ST3 :
            if      (Inp_DI=='b10) State_DN = ST2;
            else if (Inp_DI=='b01) begin
              Oup_DO   = FULL;
              State_DN = ST4;
            end
         ST4 :
            if      (Inp_DI=='b10) State_DN = ST3;
            else Oup_DO   = FULL;
         default : begin
            Oup_DO   = HIMP;
            State_DN = ST2;
         end
      endcase
   end

   // updating of state
   // ------------------------------------------------------------------------
   always_ff @(posedge Clk_CI, negedge Rst_RBI)   // sensitivity list
      // activities triggered by asynchronous reset (active low)
      if (~Rst_RBI)
         State_DP <= ST2;
      // activities triggered by rising edge of clock
      else
         State_DP <= State_DN;

endmodule
