-- Mission: Illustrate how to model a Mealy machine with two processes.
--    Example designed to include an asynchronous reset, three-state outputs,
--    self loops, symbolic encoding of states and outputs, plus handling of
--    parasitic states and inputs so as to eliminate any chance for lockup.
-- Functionality: See state graph below.     
-- Author: H.Kaeslin.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
-------------------------------------------------------------------------------

entity mealy5st is
   port (Clk_CI  : in  std_logic;
         Rst_RBI : in  std_logic;
         Inp_DI  : in  std_logic_vector(1 downto 0);
         Oup_DO  : out std_logic_vector(1 downto 0) );  -- ternary, 11 not used
end mealy5st;

-------------------------------------------------------------------------------
-- Inp_DI / Oup_DO                                                   (State_DP)
--
--              00 /        00 /        00 /        00 /        00 /
--             EMPTY        IBTW        IBTW        IBTW        FULL
--               _           _           _           _           _
--              | |  01 /   | |  01 /   | |  01 /   | |  01 /   | |
--              | |  IBTW   | |  IBTW   | |  IBTW   | |  FULL   | |
--         ---> v / ------> v / ------> v / ------> v / ------> v / ----
--   10 / |     (0)         (1)         (2)         (3)         (4)     | 01 /
--  EMPTY  -----   <--------   <-------- ^ <--------   <--------   <----  FULL
--                    10 /        10 /   |    10 /        10 / 
--                   EMPTY        IBTW   |    IBTW        IBTW              
--                                       o
-------------------------------------------------------------------------------

architecture enumerated_state of mealy5st is
   
   type state is (ST0, ST1, ST2, ST3, ST4);  -- enumerated state type
   signal State_DP, State_DN : state;        -- present state and next state
   -- symbolic encodings of output
   constant EMPTY : std_logic_vector(1 downto 0) := "10";   -- empty
   constant IBTW  : std_logic_vector(1 downto 0) := "00";   -- in between
   constant FULL  : std_logic_vector(1 downto 0) := "01";   -- full
   constant HIMP  : std_logic_vector(1 downto 0) := "ZZ";   -- high-impedance
   
begin

   ---- computation of next state and present outputs
   ---------------------------------------------------------------------------
   p_memless : process (Inp_DI, State_DP)
   begin
      -- default assignments
      State_DN <= State_DP;   -- remain in present state 
      Oup_DO <= IBTW;         -- output "in between"
      -- nondefault transitions and outputs
      case State_DP is
         when ST0 => 
            if    Inp_DI="00" then Oup_DO <= EMPTY;                   
            elsif Inp_DI="10" then Oup_DO <= EMPTY;
            elsif Inp_DI="01" then State_DN <= ST1;
            else Oup_DO <= EMPTY;   -- parasitic input 11, treat as 00
            end if;
         when ST1 =>
            if    Inp_DI="10" then Oup_DO <= EMPTY; State_DN <= ST0;
            elsif Inp_DI="01" then State_DN <= ST2;
            end if;      
         when ST2 => 
            if    Inp_DI="10" then State_DN <= ST1;
            elsif Inp_DI="01" then State_DN <= ST3;
            end if;
         when ST3 =>
            if    Inp_DI="10" then State_DN <= ST2;
            elsif Inp_DI="01" then Oup_DO <= FULL; State_DN <= ST4;
            end if;
         when ST4 =>                   
            if    Inp_DI="10" then State_DN <= ST3;
            else  Oup_DO <= FULL;   -- 01, 00 or parasitic input 11
            end if;
         when others =>   -- tie up parasitic states for synthesis
            Oup_DO <= HIMP; State_DN <= ST2; 
      end case;
   end process p_memless;

   ---- updating of state
   ----------------------------------------------------------------------------
   p_memzing : process (Clk_CI, Rst_RBI)
   begin
      -- activities triggered by asynchronous reset (active low)
      if Rst_RBI = '0' then
         State_DP <= ST2;
      -- activities triggered by rising edge of clock
      elsif Clk_CI'event and Clk_CI = '1' then
         State_DP <= State_DN;
      end if;
   end process p_memzing;

end enumerated_state;
