// Mission: Illustrate numerous ways for capturing combinational logic.
//    Example includes module bodies that follow behavioral, dataflow, 
//    and structural coding styles. In an attempt to demonstrate the richness
//    of SystemVerilog, a broad variety of coding styles has been covered. 
//    Most of them are inappropriate if data word width were allowed to vary. 
//    In fact, itera... are the only architecture bodies that do scale well.
//    Nevertheless, it is interesting to observe that today's synthesis tools 
//    produce almost the same logic network irrespective of the coding style 
//    chosen (in this combinational example). This has not always been so. 
// Functionality: so-called (4,3)-counter, see below for a truth table. 
// Findings in terms of hardware costs after unconstrained optimization
//    for UMCL65 library with SynopsysDC:
//    architecture      cells  types  nets  area    tool version
//    condassign        8      6      12    21.96   2013.12
//    concurassign      7      5      11    21.24    "
//    caseselect        8      6      12    21.96    "
//    ifelsifelse       8      6      12    21.96    "
//    lookup            7      5      11    21.60    "
//    iteracount        7      5      11    21.60    "
//    iteralogic        7      5      11    21.60    "
//    iterapad          7      5      11    21.24    "
//    dataflow          7      5      11    21.60    "
//    structgtech       8      7      12    20.16    "
// Author: B.Muheim.
//-----------------------------------------------------------------------------

`include "../sourcecode/onescntdefs.sv"

module onescnt 
   ( output logic [2:0] Cnt_DO, 
     input  logic [3:0] Inp_DI );
   //--------------------------------------------------------------------------
   //                     Cnt_DO       Cnt_DO[2]      Cnt_DO[1]      Cnt_DO[0]
   //    Inp_DI[1:0]   00 01 11 10    00 01 11 10    00 01 11 10    00 01 11 10
   // Inp_DI[3:2] 00    0  1  2  1     0  0  0  0     0  0  1  0     0  1  0  1
   //   "    "    01    1  2  3  2     0  0  0  0     0  1  1  1     1  0  1  0
   //   "    "    11    2  3  4  3     0  0  1  0     1  1  0  1     0  1  0  1
   //   "    "    10    1  2  3  2     0  0  0  0     0  1  1  1     1  0  1  0
   //--------------------------------------------------------------------------

   //==========================================================================
   `ifdef USE_CONDASSIGN     
   // mapping specified by nested conditions in a continuous assignment

       assign Cnt_DO = (Inp_DI == 4'b0000) ? 3'b000 :
                       (Inp_DI == 4'b0001 | Inp_DI == 4'b0010 | 
                        Inp_DI == 4'b0100 | Inp_DI == 4'b1000) ? 3'b001 :
                       (Inp_DI == 4'b0111 | Inp_DI == 4'b1011 |
                        Inp_DI == 4'b1101 | Inp_DI == 4'b1110) ? 3'b011 :
                       (Inp_DI == 4'b1111) ? 3'b100 :
                       3'b010;

   //==========================================================================
   `elsif USE_CONCURASSIGN     
   // mapping specified by arithmetic computation in a continuous assignment
   
       assign Cnt_DO = Inp_DI[3] + 
                       Inp_DI[2] + 
                       Inp_DI[1] + 
                       Inp_DI[0];

   //==========================================================================
   `elsif USE_CASESELECT     
   // mapping specified by exhaustive enumeration in an always_comb block 

   always_comb
      case ( Inp_DI )
      // assign the output signal the values listed below
         4'b0000 : Cnt_DO = 3'b000;
         4'b0001 : Cnt_DO = 3'b001;
         4'b0010 : Cnt_DO = 3'b001;
         4'b0011 : Cnt_DO = 3'b010;
         4'b0100 : Cnt_DO = 3'b001;
         4'b0101 : Cnt_DO = 3'b010;
         4'b0110 : Cnt_DO = 3'b010;
         4'b0111 : Cnt_DO = 3'b011;
         4'b1000 : Cnt_DO = 3'b001;
         4'b1001 : Cnt_DO = 3'b010;
         4'b1010 : Cnt_DO = 3'b010;
         4'b1011 : Cnt_DO = 3'b011;
         4'b1100 : Cnt_DO = 3'b010;
         4'b1101 : Cnt_DO = 3'b011;
         4'b1110 : Cnt_DO = 3'b011;
         4'b1111 : Cnt_DO = 3'b100;
      endcase

   //==========================================================================
   `elsif USE_IFELSIFELSE     
   // mapping specified by telescoped branching in an always_comb block 

   always_comb
   begin
      // assign the output signals the values listed below
      if (Inp_DI == 4'b0000)
         Cnt_DO = 3'b000; 
      else if (Inp_DI == 4'b0001 || Inp_DI == 4'b0010 || 
               Inp_DI == 4'b0100 || Inp_DI == 4'b1000)
         Cnt_DO = 3'b001; 
      else if (Inp_DI == 4'b0111 || Inp_DI == 4'b1011 ||
               Inp_DI == 4'b1101 || Inp_DI == 4'b1110)
         Cnt_DO = 3'b011; 
      else if (Inp_DI == 4'b1111)
         Cnt_DO = 3'b100; 
      else
         Cnt_DO = 3'b010; 

   end

   //==========================================================================
   `elsif USE_LOOKUP     
   // mapping specified by indexing a lookup table in an always_comb block 

   always_comb
   begin
      const logic [2:0] LOOKUP_TABLE[0:15] = '{0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4};
 
      // use input as index, look up in table, and assign to output
      Cnt_DO = LOOKUP_TABLE[Inp_DI];

   end
   
   //==========================================================================
   `elsif USE_ITERACOUNT     
   // mapping specified by iterative counting in an always_comb block 

   always_comb
   begin

      integer num;
      num  = 0;
      
      // count those bits that equal one in a sequential loop
      for (int idx = 0; idx<=3; idx++)
         if (Inp_DI[idx] == 1'b1) num++;

      // convert the resulting number into a binary coded signal
      Cnt_DO = num;

   end

   //==========================================================================
   `elsif USE_ITERALOGIC     
   // mapping specified by iterative logic operations in an always_comb block 
   // after K.C. Chang, Digital Systems Design with VHDL and Synthesis, p121

   always_comb
   begin
      logic [2:0] tmp;
      logic cin, cout;
      
      tmp = 3'b000;
      // increment binary sum bit by bit in a sequential loop      
      for (int idx = 0; idx<$size(Inp_DI); idx++) begin
         if (Inp_DI[idx] == 1'b1) cin = 1'b1;
         for (int idk = 0; idk<$size(tmp); idk++) begin
            cout = tmp[idk] & cin;
            tmp[idk] = tmp[idk] ^ cin;
            cin = cout;
         end
      end

      Cnt_DO = tmp;

   end

   //==========================================================================
   `elsif USE_ITERAPAD     
   // mapping specified by iterative arithmetic computation in an always_comb block 

   // note: the basic approach is the same as in USE_CONCURASSIGN, 
   // yet the embodiment as an always_comb block  permits the usage 
   // of an iterative loop and of local variables.

   always_comb
   begin
 
      integer sum;    
      sum  = 0;

      // pad input bits with zeroes before adding them in a sequential loop   
      for (int idx = 0; idx<$size(Inp_DI); idx++)
         sum = sum + Inp_DI[idx];

      Cnt_DO = sum;

   end

   //==========================================================================
   `elsif USE_DATAFLOW     
   // mapping specified by continuous assignments at the Boolean level

   logic Row0_D, Row3_D, Row1or2_D, Col3_D, Col1thru3_D;

   // indicate a combinational function for each output bit
   // by exploiting the similarities of rows in the truth table
   assign Row0_D = ~(Inp_DI[3] | Inp_DI[2]);
   assign Row3_D = Inp_DI[3] & Inp_DI[2];
   assign Row1or2_D = ~Row0_D & ~Row3_D;
   assign Col3_D = Inp_DI[1] & Inp_DI[0];
   assign Col1thru3_D = Inp_DI[1] | Inp_DI[0];
   assign Cnt_DO[2] = Row3_D & Col3_D;
   assign Cnt_DO[1] = (Row0_D & Col3_D) | (Row3_D & ~Col3_D) |
                      (Row1or2_D & Col1thru3_D);
   assign Cnt_DO[0] = Inp_DI[3] ^ Inp_DI[2] ^ Inp_DI[1] ^ Inp_DI[0];
   
   //==========================================================================
   `elsif USE_STRUCTGTECH     
   // logic network described as a bunch of interconnected logic gates

   // declare internal signals
   logic Yellow_D, Green_D, Blue_D, Violet_D;

   // instantiate components and connect them by listing port maps
   GTECH_OA21 u1 ( .A(Inp_DI[3]), .B(Blue_D), .C(Green_D), .Z(Cnt_DO[0]) );
   GTECH_NOR2 u2 ( .A(Green_D), .B(Violet_D), .Z(Cnt_DO[2]) );
   GTECH_XOR2 u3 ( .A(Inp_DI[1]), .B(Inp_DI[0]), .Z(Yellow_D) );
   GTECH_XOR2 u4 ( .A(Yellow_D), .B(Inp_DI[2]), .Z(Blue_D) );
   GTECH_XOR2 u5 ( .A(Green_D), .B(Violet_D), .Z(Cnt_DO[1]) );
   GTECH_NAND2 u6 ( .A(Inp_DI[3]), .B(Blue_D), .Z(Green_D) );
   GTECH_AOI22 u7 ( .A(Inp_DI[0]), .B(Inp_DI[1]), .C(Yellow_D), .D(Inp_DI[2]),
                    .Z(Violet_D) );

   `endif

endmodule

