-- Mission: Illustrate numerous ways for capturing combinational logic.
--    Example includes architecture bodies that follow behavioral, dataflow,
--    and structural coding styles. In an attempt to demonstrate the richness
--    of VHDL, a broad variety of coding styles has been covered. 
--    Most of them are inappropriate if data word width were allowed to vary.
--    In fact, itera... are the only architecture bodies that do scale well. 
--    Nevertheless, it is interesting to observe that today's synthesis tools 
--    produce almost the same logic network irrespective of the coding style 
--    chosen (in this combinational example). This has not always been so. 
-- Functionality: so-called (4,3)-counter, see below for a truth table.
-- Findings in terms of hardware costs after unconstrained optimization
--    for UMCL65 library with SynopsysDC:
--    architecture      cells  types  nets  area    tool version
--    selectassign      7      5      11    21.60   2013.12
--    condassign        7      5      11    21.60    "
--    concurassign      7      5      11    21.24    "
--    caseselect        7      5      11    21.60    "
--    ifelsifelse       7      5      11    21.60    "
--    lookup            7      5      11    21.60    "
--    iteracount        7      5      11    21.24    "
--    iteralogic        7      5      11    21.60    "
--    iterapad          7      5      11    21.24    "
--    dataflow          7      5      11    21.60    "
--    structgtech       8      7      12    20.16    "
-- Author: H.Kaeslin.

-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library gtech;                  -- generic component library by Synopsys
use gtech.gtech_components.all;
-------------------------------------------------------------------------------

entity onescnt is
   port ( Inp_DI : in  std_logic_vector(3 downto 0);
          Cnt_DO : out std_logic_vector(2 downto 0) );
end onescnt;

-------------------------------------------------------------------------------
--                     Cnt_DO       Cnt_DO(2)      Cnt_DO(1)      Cnt_DO(0)
--    Inp_DI(1:0)   00 01 11 10    00 01 11 10    00 01 11 10    00 01 11 10
-- Inp_DI(3:2) 00    0  1  2  1     0  0  0  0     0  0  1  0     0  1  0  1
--   "    "    01    1  2  3  2     0  0  0  0     0  1  1  1     1  0  1  0
--   "    "    11    2  3  4  3     0  0  1  0     1  1  0  1     0  1  0  1
--   "    "    10    1  2  3  2     0  0  0  0     0  1  1  1     1  0  1  0
-------------------------------------------------------------------------------

--=============================================================================

-- mapping specified by exhaustive enumeration in a selected signal assignment
architecture selectassign of onescnt is
begin
   -- assign the output signal the values listed below   
   with Inp_DI select
      Cnt_DO <= "000" when "0000",
                "001" when "0001",
                "001" when "0010",
                "001" when "0100",
                "001" when "1000",
                "011" when "0111",
                "011" when "1011",
                "011" when "1101",
                "011" when "1110",
                "100" when "1111",
                "010" when others;
end selectassign;

--=============================================================================

-- mapping specified by nested conditions in a conditional signal assignment
architecture condassign of onescnt is
begin

   Cnt_DO <= "000" when Inp_DI = "0000" else 
             "001" when Inp_DI = "0001" or Inp_DI = "0010" or
                        Inp_DI = "0100" or Inp_DI = "1000" else
             "011" when Inp_DI = "0111" or Inp_DI = "1011" or
                        Inp_DI = "1101" or Inp_DI = "1110" else
             "100" when Inp_DI = "1111" else
             "010";
end condassign;

--=============================================================================

-- mapping specified by arithmetic computation in a concurrent signal assignment
architecture concurassign of onescnt is
begin
   -- extend input bits to final word width before adding them
   
   -- note: [type-]qualified expressions make it possible to interpret
   -- the result of concatenation as unsigned without explicit conversion;

   Cnt_DO <= std_logic_vector( unsigned'("00" & Inp_DI(3)) +
                               unsigned'("00" & Inp_DI(2)) +
                               unsigned'("00" & Inp_DI(1)) +
                               unsigned'("00" & Inp_DI(0)) );
end concurassign;

--=============================================================================

-- mapping specified by exhaustive enumeration in a process statement
architecture caseselect of onescnt is
begin
   p_memless : process (Inp_DI)
   begin
      -- assign the output signal the values listed below
      case Inp_DI is
         when "0000" => Cnt_DO<="000";
         when "0001" => Cnt_DO<="001";
         when "0010" => Cnt_DO<="001";
         when "0011" => Cnt_DO<="010";
         when "0100" => Cnt_DO<="001";
         when "0101" => Cnt_DO<="010";
         when "0110" => Cnt_DO<="010";
         when "0111" => Cnt_DO<="011";
         when "1000" => Cnt_DO<="001";
         when "1001" => Cnt_DO<="010";
         when "1010" => Cnt_DO<="010";
         when "1011" => Cnt_DO<="011";
         when "1100" => Cnt_DO<="010";
         when "1101" => Cnt_DO<="011";
         when "1110" => Cnt_DO<="011";
         when "1111" => Cnt_DO<="100";
         when others => Cnt_DO<="111";  -- SynopsysDC insists on this
      end case;
   end process p_memless;
end caseselect;

--=============================================================================

-- mapping specified by telescoped branching in a process statement
architecture ifelsifelse of onescnt is
begin
   p_memless : process (Inp_DI)
   begin
      -- assign the output signals the values listed below
      if Inp_DI="0000" then
         Cnt_DO<="000";
      elsif (Inp_DI="0001" or Inp_DI="0010" or Inp_DI="0100" or Inp_DI="1000") 
      then
         Cnt_DO<="001";
      elsif (Inp_DI="0111" or Inp_DI="1011" or Inp_DI="1101" or Inp_DI="1110") 
      then
         Cnt_DO<="011";    
      elsif Inp_DI="1111" then
         Cnt_DO<="100";
      else
         Cnt_DO<="010";
      end if;
   end process p_memless;
end ifelsifelse;

--=============================================================================

-- mapping specified by indexing a lookup table in a process statement
architecture lookup of onescnt is 
begin
   -- address of array is requested to be of type integer or natural
   p_memless : process (Inp_DI)
      variable num : natural range 0 to 4;
      variable address : natural range 0 to 15;
      type table_16 is array(0 to 15) of natural range 0 to 4;
      constant LOOKUP_TABLE: table_16 := (0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4);
   begin
      -- use input as index, look up in table, and assign to output
      address := to_integer(unsigned(Inp_DI));    
      num := LOOKUP_TABLE(address);
      Cnt_DO <= std_logic_vector(to_unsigned(num,3));     
   end process p_memless;
end lookup;

--=============================================================================

-- mapping specified by iterative counting in a process statement
architecture iteracount of onescnt is
begin
   p_memless : process (Inp_DI)
      variable num : natural range 0 to 4;
   begin
      num := 0;  
      -- count those bits that equal one in a sequential loop
      for idx in 0 to 3 loop
         if Inp_DI(idx)='1' then num := num+1; end if;
      end loop;
      -- convert the resulting number into a binary coded signal
      Cnt_DO <= std_logic_vector(to_unsigned(num,3));
   end process p_memless;
end iteracount;

--=============================================================================

-- mapping specified by iterative logic operations in a process statement
-- after K.C. Chang, Digital Systems Design with VHDL and Synthesis, p121
architecture iteralogic of onescnt is
begin
   p_memless : process (Inp_DI)
      variable tmp : std_logic_vector(2 downto 0);
      variable cin, cout : std_logic;
   begin
      tmp := "000";
      -- increment binary sum bit by bit in a sequential loop      
      for idx in Inp_DI'reverse_range loop
         if Inp_DI(idx)='1' then
            cin := '1';
            -- propagate carries
            for idk in tmp'reverse_range loop
               cout := tmp(idk) and cin;
               tmp(idk) := tmp(idk) xor cin;
               cin := cout;
            end loop;
         end if;
      end loop;
      Cnt_DO <= tmp;
   end process p_memless;
end iteralogic;

--=============================================================================

-- mapping specified by iterative arithmetic computation in a process statement

-- note: the basic approach is the same as in architecture body concurassign, 
-- yet the embodiment as a process permits the usage of an iterative loop
-- and of variables.

architecture iterapad of onescnt is
begin
   p_memless : process (Inp_DI)
      variable pad, sum : unsigned(2 downto 0);
   begin
      sum := (others => '0');   -- with-independent shorthand
      -- pad input bits with zeroes before adding them in a sequential loop   
      for idx in Inp_DI'range loop
         pad := "00" & Inp_DI(idx); 
         sum := sum + pad;
      end loop;
      Cnt_DO <= std_logic_vector(sum);
   end process p_memless;
end iterapad;

--=============================================================================

-- mapping specified by concurrent signal assignments at the Boolean level
architecture dataflow of onescnt is
   signal Row0_D, Row3_D, Row1or2_D, Col3_D, Col1thru3_D: std_logic;
begin
   -- indicate a combinational function for each output bit
   -- by exploiting the similarities of rows in the truth table
   Row0_D <= Inp_DI(3) nor Inp_DI(2);
   Row3_D <= Inp_DI(3) and Inp_DI(2);
   Row1or2_D <= (not Row0_D) and (not Row3_D);
   Col3_D <= Inp_DI(1) and Inp_DI(0);
   Col1thru3_D <= Inp_DI(1) or Inp_DI(0);
   Cnt_DO(2) <= Row3_D and Col3_D;
   Cnt_DO(1) <= (Row0_D and Col3_D) or (Row3_D and (not Col3_D)) or
                (Row1or2_D and Col1thru3_D);
   Cnt_DO(0) <= Inp_DI(3) xor Inp_DI(2) xor Inp_DI(1) xor Inp_DI(0);
end dataflow;

--=============================================================================

-- logic network described as a bunch of interconnected logic gates
architecture structgtech of onescnt is

   -- note: a list of explicit component declarations is redundant in this
   -- context because the GTECH generic component library has been made
   -- available by the library and use statements early in this program.

   -- declare internal signals
   signal Yellow_D, Green_D, Blue_D, Violet_D: std_logic;

begin
   
   -- instantiate components and connect them by listing port maps
   u1 : GTECH_OA21
      port map ( A=>Inp_DI(3), B=>Blue_D, C=>Green_D, Z=>Cnt_DO(0) );
   u2 : GTECH_NOR2
      port map ( A=>Green_D, B=>Violet_D, Z=>Cnt_DO(2) );
   u3 : GTECH_XOR2
      port map ( A=>Inp_DI(1), B=>Inp_DI(0), Z=>Yellow_D );
   u4 : GTECH_XOR2
      port map ( A=>Yellow_D, B=>Inp_DI(2), Z=>Blue_D );
   u5 : GTECH_XOR2
      port map ( A=>Green_D, B=>Violet_D, Z=>Cnt_DO(1) );
   u6 : GTECH_NAND2
      port map ( A=>Inp_DI(3), B=>Blue_D, Z=>Green_D );
   u7 : GTECH_AOI22
      port map ( A=>Inp_DI(0), B=>Inp_DI(1), C=>Yellow_D, D=>Inp_DI(2),
                 Z=>Violet_D );

end structgtech;
